import numpy as np
import pytest

from autofit.tools import fit


class TestDataFitter:

    def test__image_and_model_are_identical__no_masking__check_values_are_correct(self):

        data = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map = np.array([2.0, 2.0, 2.0, 2.0])
        mask = np.array([False, False, False, False])

        model_data = np.array([1.0, 2.0, 3.0, 4.0])

        data_fit = fit.DataFit(data=data, noise_map=noise_map, mask=mask, model_data=model_data)

        assert (data_fit.data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_map == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.signal_to_noise_map == np.array([0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit.mask == np.array([False, False, False, False])).all()
        assert (data_fit.model_data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.residual_map == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert (data_fit.chi_squared_map == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert data_fit.chi_squared == 0.0
        assert data_fit.reduced_chi_squared == 0.0
        assert data_fit.noise_normalization == np.sum(np.log(2 * np.pi * noise_map ** 2.0))
        assert data_fit.likelihood == -0.5*(data_fit.chi_squared + data_fit.noise_normalization)

    def test__image_and_model_mismatch__no_masking__check_values_are_correct(self):

        data = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map = np.array([2.0, 2.0, 2.0, 2.0])
        mask = np.array([False, False, False, False])

        model_data = np.array([1.0, 1.0, 1.0, 1.0])

        data_fit = fit.DataFit(data=data, noise_map=noise_map, mask=mask, model_data=model_data)

        assert (data_fit.data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_map == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.signal_to_noise_map == np.array([0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit.mask == np.array([False, False, False, False])).all()
        assert (data_fit.model_data == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (data_fit.residual_map == np.array([0.0, 1.0, 2.0, 3.0])).all()
        assert (data_fit.chi_squared_map == np.array([0.0, (1.0 / 2.0) ** 2.0, (2.0 / 2.0) ** 2.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit.chi_squared == (1.0 / 2.0) ** 2.0 + (2.0 / 2.0) ** 2.0 + (3.0 / 2.0) ** 2.0
        assert data_fit.reduced_chi_squared == data_fit.chi_squared / 4.0
        assert data_fit.noise_normalization == np.sum(np.log(2 * np.pi * noise_map ** 2.0))
        assert data_fit.likelihood == -0.5*(data_fit.chi_squared + data_fit.noise_normalization)

    def test__image_and_model_mismatch__include_masking__check_values_are_correct(self):

        data = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map = np.array([2.0, 2.0, 2.0, 2.0])
        mask = np.array([False, True, True, False])

        model_data = np.array([1.0, 1.0, 1.0, 1.0])

        data_fit = fit.DataFit(data=data, noise_map=noise_map, mask=mask, model_data=model_data)

        assert (data_fit.data == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_map == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.signal_to_noise_map == np.array([0.5, 1.0, 1.5, 2.0])).all()
        assert (data_fit.mask == np.array([False, True, True, False])).all()
        assert (data_fit.model_data == model_data).all()
        assert (data_fit.residual_map == np.array([0.0, 0.0, 0.0, 3.0])).all()
        assert (data_fit.chi_squared_map == np.array([0.0, 0.0, 0.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit.chi_squared == (3.0 / 2.0) ** 2.0
        assert data_fit.reduced_chi_squared == data_fit.chi_squared / 2.0
        assert data_fit.noise_normalization == 2.0 * np.sum(np.log(2 * np.pi * 2.0 ** 2.0))
        assert data_fit.likelihood == -0.5*(data_fit.chi_squared + data_fit.noise_normalization)


class TestDataFitterStack:

    def test__image_and_model_are_identical__no_masking__check_values_are_correct(self):

        data_0 = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map_0 = np.array([2.0, 2.0, 2.0, 2.0])
        mask_0 = np.array([False, False, False, False])
        model_data_0 = np.array([1.0, 2.0, 3.0, 4.0])

        data_1 = np.array([4.0, 3.0, 2.0, 1.0])
        noise_map_1 = np.array([3.0, 3.0, 3.0, 3.0])
        mask_1 = np.array([False, False, False, False])
        model_data_1 = np.array([4.0, 3.0, 2.0, 1.0])

        data_fit = fit.DataFitStack(datas=[data_0, data_1], noise_maps=[noise_map_0, noise_map_1],
                               masks=[mask_0, mask_1], model_datas=[model_data_0, model_data_1])

        assert (data_fit.datas[0] == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_maps[0] == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.masks[0] == np.array([False, False, False, False])).all()
        assert (data_fit.model_datas[0] == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.residual_maps[0] == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert (data_fit.chi_squared_maps[0] == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert data_fit.chi_squareds[0] == 0.0
        assert data_fit.reduced_chi_squareds[0] == 0.0
        assert data_fit.noise_normalizations[0] == np.sum(np.log(2 * np.pi * noise_map_0 ** 2.0))
        assert data_fit.likelihoods[0] == -0.5 * (data_fit.chi_squareds[0] + data_fit.noise_normalizations[0])

        assert (data_fit.datas[1] == np.array([4.0, 3.0, 2.0, 1.0])).all()
        assert (data_fit.noise_maps[1] == np.array([3.0, 3.0, 3.0, 3.0])).all()
        assert (data_fit.masks[1] == np.array([False, False, False, False])).all()
        assert (data_fit.model_datas[1] == np.array([4.0, 3.0, 2.0, 1.0])).all()
        assert (data_fit.residual_maps[1] == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert (data_fit.chi_squared_maps[1] == np.array([0.0, 0.0, 0.0, 0.0])).all()
        assert data_fit.chi_squareds[1] == 0.0
        assert data_fit.reduced_chi_squareds[1] == 0.0
        assert data_fit.noise_normalizations[1] == np.sum(np.log(2 * np.pi * noise_map_1 ** 2.0))
        assert data_fit.likelihoods[1] == -0.5 * (data_fit.chi_squareds[1] + data_fit.noise_normalizations[1])

        assert data_fit.chi_squared == 0.0
        assert data_fit.reduced_chi_squared == 0.0
        assert data_fit.noise_normalization == data_fit.noise_normalizations[0] + data_fit.noise_normalizations[1]
        assert data_fit.likelihood == -0.5 * (data_fit.chi_squared + data_fit.noise_normalization)

    def test__image_and_model_mismatch__no_masking__check_values_are_correct(self):

        data_0 = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map_0 = np.array([2.0, 2.0, 2.0, 2.0])
        mask_0 = np.array([False, False, False, False])
        model_data_0 = np.array([1.0, 1.0, 1.0, 1.0])

        data_1 = np.array([4.0, 3.0, 2.0, 1.0])
        noise_map_1 = np.array([3.0, 3.0, 3.0, 3.0])
        mask_1 = np.array([False, False, False, False])
        model_data_1 = np.array([2.0, 2.0, 2.0, 2.0])

        data_fit = fit.DataFitStack(datas=[data_0, data_1], noise_maps=[noise_map_0, noise_map_1],
                               masks=[mask_0, mask_1], model_datas=[model_data_0, model_data_1])

        assert (data_fit.datas[0] == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_maps[0] == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.masks[0] == np.array([False, False, False, False])).all()
        assert (data_fit.model_datas[0] == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (data_fit.residual_maps[0] == np.array([0.0, 1.0, 2.0, 3.0])).all()
        assert (data_fit.chi_squared_maps[0] == np.array([0.0, (1.0 / 2.0) ** 2.0, (2.0 / 2.0) ** 2.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit.chi_squareds[0] == (1.0 / 2.0) ** 2.0 + (2.0 / 2.0) ** 2.0 + (3.0 / 2.0) ** 2.0
        assert data_fit.reduced_chi_squareds[0] == data_fit.chi_squareds[0] / 4.0
        assert data_fit.noise_normalizations[0] == np.sum(np.log(2 * np.pi * noise_map_0 ** 2.0))
        assert data_fit.likelihoods[0] == -0.5 * (data_fit.chi_squareds[0] + data_fit.noise_normalizations[0])

        assert (data_fit.datas[1] == np.array([4.0, 3.0, 2.0, 1.0])).all()
        assert (data_fit.noise_maps[1] == np.array([3.0, 3.0, 3.0, 3.0])).all()
        assert (data_fit.masks[1] == np.array([False, False, False, False])).all()
        assert (data_fit.model_datas[1] == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.residual_maps[1] == np.array([2.0, 1.0, 0.0, -1.0])).all()
        assert (data_fit.chi_squared_maps[1] == np.array([(2.0 / 3.0) ** 2.0, (1.0 / 3.0) ** 2.0, 0.0, (-1.0 / 3.0) ** 2.0])).all()
        assert data_fit.chi_squareds[1] == (2.0 / 3.0) ** 2.0 + (1.0 / 3.0) ** 2.0 + (-1.0 / 3.0) ** 2.0
        assert data_fit.reduced_chi_squareds[1] == data_fit.chi_squareds[1] / 4.0
        assert data_fit.noise_normalizations[1] == np.sum(np.log(2 * np.pi * noise_map_1 ** 2.0))
        assert data_fit.likelihoods[1] == -0.5 * (data_fit.chi_squareds[1] + data_fit.noise_normalizations[1])

        assert data_fit.chi_squared == data_fit.chi_squareds[0] + data_fit.chi_squareds[1]
        assert data_fit.reduced_chi_squared == data_fit.reduced_chi_squareds[0] + data_fit.reduced_chi_squareds[1]
        assert data_fit.noise_normalization == data_fit.noise_normalizations[0] + data_fit.noise_normalizations[1]
        assert data_fit.likelihood == -0.5 * (data_fit.chi_squared + data_fit.noise_normalization)

    def test__image_and_model_mismatch__include_masking__check_values_are_correct(self):

        data_0 = np.array([1.0, 2.0, 3.0, 4.0])
        noise_map_0 = np.array([2.0, 2.0, 2.0, 2.0])
        mask_0 = np.array([False, True, True, False])
        model_data_0 = np.array([1.0, 1.0, 1.0, 1.0])

        data_1 = np.array([4.0, 3.0, 2.0, 1.0])
        noise_map_1 = np.array([3.0, 3.0, 3.0, 3.0])
        mask_1 = np.array([True, True, False, False])
        model_data_1 = np.array([2.0, 2.0, 2.0, 2.0])

        data_fit = fit.DataFitStack(datas=[data_0, data_1], noise_maps=[noise_map_0, noise_map_1],
                               masks=[mask_0, mask_1], model_datas=[model_data_0, model_data_1])

        assert (data_fit.datas[0] == np.array([1.0, 2.0, 3.0, 4.0])).all()
        assert (data_fit.noise_maps[0] == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.masks[0] == np.array([False, True, True, False])).all()
        assert (data_fit.model_datas[0] == np.array([1.0, 1.0, 1.0, 1.0])).all()
        assert (data_fit.residual_maps[0] == np.array([0.0, 0.0, 0.0, 3.0])).all()
        assert (data_fit.chi_squared_maps[0] == np.array([0.0, 0.0, 0.0, (3.0 / 2.0) ** 2.0])).all()
        assert data_fit.chi_squareds[0] == (3.0 / 2.0) ** 2.0
        assert data_fit.reduced_chi_squareds[0] == data_fit.chi_squareds[0] / 2.0
        assert data_fit.noise_normalizations[0] == 2.0 * np.sum(np.log(2 * np.pi * 2.0 ** 2.0))
        assert data_fit.likelihoods[0] == -0.5 * (data_fit.chi_squareds[0] + data_fit.noise_normalizations[0])

        assert (data_fit.datas[1] == np.array([4.0, 3.0, 2.0, 1.0])).all()
        assert (data_fit.noise_maps[1] == np.array([3.0, 3.0, 3.0, 3.0])).all()
        assert (data_fit.masks[1] == np.array([True, True, False, False])).all()
        assert (data_fit.model_datas[1] == np.array([2.0, 2.0, 2.0, 2.0])).all()
        assert (data_fit.residual_maps[1] == np.array([0.0, 0.0, 0.0, -1.0])).all()
        assert (data_fit.chi_squared_maps[1] == np.array([0.0, 0.0, 0.0, (-1.0 / 3.0) ** 2.0])).all()
        assert data_fit.chi_squareds[1] == (-1.0 / 3.0) ** 2.0
        assert data_fit.reduced_chi_squareds[1] == data_fit.chi_squareds[1] / 2.0
        assert data_fit.noise_normalizations[1] == 2.0 * np.sum(np.log(2 * np.pi * 3.0 ** 2.0))
        assert data_fit.likelihoods[1] == -0.5 * (data_fit.chi_squareds[1] + data_fit.noise_normalizations[1])

        assert data_fit.chi_squared == data_fit.chi_squareds[0] + data_fit.chi_squareds[1]
        assert data_fit.reduced_chi_squared == data_fit.reduced_chi_squareds[0] + data_fit.reduced_chi_squareds[1]
        assert data_fit.noise_normalization == data_fit.noise_normalizations[0] + data_fit.noise_normalizations[1]
        assert data_fit.likelihood == pytest.approx(-0.5 * (data_fit.chi_squared + data_fit.noise_normalization), 1.0e4)


import copy
import inspect
import itertools
import re

from autofit import conf, exc
from autofit.mapper.prior import cast_collection, PriorNameValue, ConstantNameValue, TuplePrior, UniformPrior, \
    LogUniformPrior, GaussianPrior, Constant, Prior, AttributeNameValue


def tuple_name(attribute_name):
    """
    Extract the name of a tuple attribute from the name of one of its components, e.g. centre_0 -> origin

    Parameters
    ----------
    attribute_name: str
        The name of an attribute which is a component of a tuple

    Returns
    -------
    tuple_name: str
        The name of the tuple of which the attribute is a member
    """
    return "_".join(attribute_name.split("_")[:-1])


def is_tuple_like_attribute_name(attribute_name):
    """
    Determine if a string matches the pattern "{attribute_name}_#", that is if it seems to be a tuple.

    Parameters
    ----------
    attribute_name: str
        The name of some attribute that may refer to a tuple.

    Returns
    -------
    is_tuple_like: bool
        True iff the attribute name looks like that which refers to a tuple.
    """
    pattern = re.compile("^[a-zA-Z_0-9]*_[0-9]$")
    return pattern.match(attribute_name)


class PriorModelNameValue(AttributeNameValue):
    @property
    def prior_model(self):
        return self.value


class AbstractPriorModel:
    """
    Abstract model that maps a set of priors to a particular class. Must be overridden by any prior model so that the \
    model mapper recognises its prior model attributes.
    """
    _ids = itertools.count()

    @property
    def flat_prior_model_tuples(self):
        """
        Returns
        -------
        prior_models: [(str, AbstractPriorModel)]
            A list of prior models associated with this instance
        """
        raise NotImplementedError("PriorModels must implement the flat_prior_models property")

    @property
    def prior_tuples(self):
        raise NotImplementedError()

    @property
    @cast_collection(PriorModelNameValue)
    def direct_prior_model_tuples(self):
        return [(name, value) for name, value in self.__dict__.items() if isinstance(value, AbstractPriorModel)]

    @property
    def constant_tuples(self):
        raise NotImplementedError()

    @property
    def prior_class_dict(self):
        raise NotImplementedError()

    def instance_for_arguments(self, arguments):
        raise NotImplementedError()

    def name_for_prior(self, prior):
        for prior_model_name, prior_model in self.direct_prior_model_tuples:
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(prior_model_name, prior_name)
        prior_tuples = self.prior_tuples
        for name, p in prior_tuples:
            if p == prior:
                return name


class ListPriorModel(list, AbstractPriorModel):
    def name_for_prior(self, prior):
        for i, prior_model in enumerate(self):
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(i, prior_name)

    @property
    def flat_prior_model_tuples(self):
        return [flat_prior_model for prior_model in self for flat_prior_model in prior_model.flat_prior_model_tuples]

    def __init__(self, prior_models):
        """
        A prior model used to represent a list of prior models for convenience.

        Parameters
        ----------
        prior_models: [PriorModel]
            A list of prior models
        """
        self.component_number = next(self._ids)
        super().__init__(prior_models)

    @property
    @cast_collection(PriorModelNameValue)
    def label_prior_model_tuples(self):
        return [(prior_model.mapping_name if hasattr(prior_model, "mapping_name") else str(i), prior_model) for
                i, prior_model in enumerate(self)]

    def instance_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        model_instances: [object]
            A list of instances constructed from the list of prior models.
        """
        return [prior_model.instance_for_arguments(arguments) for prior_model in self]

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        prior_models: [PriorModel]
            A new list of prior models with gaussian priors
        """
        return ListPriorModel(
            [prior_model.gaussian_prior_model_for_arguments(arguments) for prior_model in self])

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Union(Prior, TuplePrior))]
        """
        return set([prior for prior_model in self for prior in prior_model.prior_tuples])

    @property
    @cast_collection(ConstantNameValue)
    def constant_tuples(self):
        """
        Returns
        -------
        priors: [(String, Union(Prior, TuplePrior))]
        """
        return set([constant for prior_model in self for constant in prior_model.constant_tuples])

    @property
    def prior_class_dict(self):
        return {prior: cls for prior_model in self for prior, cls in prior_model.prior_class_dict.items()}


class PriorModel(AbstractPriorModel):
    """Object comprising class and associated priors
        @DynamicAttrs
    """

    @property
    def flat_prior_model_tuples(self):
        return [("", self)]

    def __init__(self, cls):
        """
        Parameters
        ----------
        cls: class
            The class associated with this instance
        """

        self.cls = cls
        self.component_number = next(self._ids)

        arg_spec = inspect.getfullargspec(cls.__init__)

        try:
            defaults = dict(zip(arg_spec.args[-len(arg_spec.defaults):], arg_spec.defaults))
        except TypeError:
            defaults = {}

        args = arg_spec.args[1:]

        if 'settings' in defaults:
            del defaults['settings']
        if 'settings' in args:
            args.remove('settings')

        for arg in args:
            if arg in defaults and isinstance(defaults[arg], tuple):
                tuple_prior = TuplePrior()
                for i in range(len(defaults[arg])):
                    attribute_name = "{}_{}".format(arg, i)
                    setattr(tuple_prior, attribute_name, self.make_prior(attribute_name, cls))
                setattr(self, arg, tuple_prior)
            else:
                setattr(self, arg, self.make_prior(arg, cls))

    @staticmethod
    def make_prior(attribute_name, cls):
        """
        Create a prior for an attribute of a class with a given name. The prior is created by searching the default
        prior config for the attribute.

        Entries in configuration with a u become uniform priors; with a g become gaussian priors; with a c become
        constants.

        If prior configuration for a given attribute is not specified in the configuration for a class then the
        configuration corresponding to the parents of that class is searched. If no configuration can be found then a
        prior exception is raised.

        Parameters
        ----------
        attribute_name: str
            The name of the attribute for which a prior is created
        cls: class
            The class to which the attribute belongs

        Returns
        -------
        prior: p.Prior
            A prior

        Raises
        ------
        exc.PriorException
            If no configuration can be found
        """
        config_arr = conf.instance.prior_default.get_for_nearest_ancestor(cls, attribute_name)
        if config_arr[0] == "u":
            return UniformPrior(config_arr[1], config_arr[2])
        elif config_arr[0] == "l":
            return LogUniformPrior(config_arr[1], config_arr[2])
        elif config_arr[0] == "g":
            limits = conf.instance.prior_limit.get_for_nearest_ancestor(cls, attribute_name)
            return GaussianPrior(config_arr[1], config_arr[2], *limits)
        elif config_arr[0] == "c":
            return Constant(config_arr[1])
        raise exc.PriorException(
            "Default prior for {} has no type indicator (u - Uniform, g - Gaussian, c - Constant".format(
                attribute_name))

    def linked_model_for_class(self, cls, make_constants_variable=False, **kwargs):
        """
        Create a PriorModel wrapping the specified class with attributes from this instance. Priors can be overridden
        using keyword arguments. Any constructor arguments of the new class for which there is no attribute associated
        with this class and no keyword argument are created from config.

        If make_constants_variable is True then constants associated with this instance will be used to set the mean
        of priors in the new instance rather than overriding them.

        Parameters
        ----------
        cls: class
            The class that the new PriorModel will wrap
        make_constants_variable: bool
            If True constants from this instance will be used to determine the mean values for priors in the new
            instance rather than overriding them
        kwargs
            Keyword arguments passed in here are used to override attributes from this instance or add new attributes

        Returns
        -------
        new_model: PriorModel
            A new prior model with priors derived from this instance
        """
        constructor_args = inspect.getfullargspec(cls).args
        attribute_tuples = self.attribute_tuples
        new_model = PriorModel(cls)
        for attribute_tuple in attribute_tuples:
            name = attribute_tuple.name
            if name in constructor_args or (
                    is_tuple_like_attribute_name(name) and tuple_name(name) in constructor_args):
                attribute = kwargs[name] if name in kwargs else attribute_tuple.value
                if make_constants_variable and isinstance(attribute, Constant):
                    new_attribute = getattr(new_model, name)
                    if isinstance(new_attribute, Prior):
                        new_attribute.mean = attribute.value
                        continue
                setattr(new_model, name, attribute)
        return new_model

    def __setattr__(self, key, value):
        if key not in ("component_number", "phase_property_position", "mapping_name"):
            try:
                if "_" in key:
                    name = key.split("_")[0]
                    tuple_prior = [v for k, v in self.tuple_prior_tuples if name == k][0]
                    setattr(tuple_prior, key, value)
                    return
            except IndexError:
                pass
            if isinstance(value, float) or isinstance(value, int):
                super().__setattr__(key, Constant(value))
                return
        super(PriorModel, self).__setattr__(key, value)

    def __getattr__(self, item):
        try:
            if "_" in item:
                return getattr([v for k, v in self.tuple_prior_tuples if item.split("_")[0] == k][0], item)

        except IndexError:
            pass
        self.__getattribute__(item)

    @property
    @cast_collection(PriorNameValue)
    def tuple_prior_tuples(self):
        """
        Returns
        -------
        tuple_prior_tuples: [(String, TuplePrior)]
        """
        return list(filter(lambda t: type(t[1]) is TuplePrior, self.__dict__.items()))

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        """
        Returns
        -------
        direct_priors: [(String, Prior)]
        """
        return list(filter(lambda t: isinstance(t[1], Prior), self.__dict__.items()))

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Prior))]
        """
        return [prior for tuple_prior in self.tuple_prior_tuples for prior in
                tuple_prior[1].prior_tuples] + self.direct_prior_tuples

    @property
    @cast_collection(ConstantNameValue)
    def direct_constant_tuples(self):
        """
        Returns
        -------
        constants: [(String, Constant)]
            A list of constants
        """
        return list(filter(lambda t: isinstance(t[1], Constant), self.__dict__.items()))

    @property
    @cast_collection(ConstantNameValue)
    def constant_tuples(self):
        """
        Returns
        -------
        constants: [(String, Constant)]
        """
        return [constant_tuple for tuple_prior in self.tuple_prior_tuples for constant_tuple in
                tuple_prior[1].constant_tuples] + self.direct_constant_tuples

    @property
    @cast_collection(AttributeNameValue)
    def attribute_tuples(self):
        return self.prior_tuples + self.constant_tuples

    @property
    def prior_class_dict(self):
        return {prior[1]: self.cls for prior in self.prior_tuples}

    def instance_for_arguments(self, arguments: {Prior: float}):
        """
        Create an instance of the associated class for a set of arguments

        Parameters
        ----------
        arguments: {Prior: float}
            Dictionary mapping_matrix priors to attribute analysis_path and value pairs

        Returns
        -------
            An instance of the class
        """
        for prior, value in arguments.items():
            prior.assert_within_limits(value)
        model_arguments = {t.name: arguments[t.prior] for t in self.direct_prior_tuples}
        constant_arguments = {t.name: t.constant.value for t in self.direct_constant_tuples}
        for tuple_prior in self.tuple_prior_tuples:
            model_arguments[tuple_prior.name] = tuple_prior.prior.value_for_arguments(arguments)

        return self.cls(**{**model_arguments, **constant_arguments})

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Create a new instance of model mapper with a set of Gaussian priors based on tuples provided by a previous \
        nonlinear search.

        Parameters
        ----------
        arguments: [(float, float)]
            Tuples providing the mean and sigma of gaussians

        Returns
        -------
        new_model: ModelMapper
            A new model mapper populated with Gaussian priors
        """
        new_model = copy.deepcopy(self)

        model_arguments = {t.name: arguments[t.prior] for t in self.direct_prior_tuples}

        for tuple_prior_tuple in self.tuple_prior_tuples:
            setattr(new_model, tuple_prior_tuple.name,
                    tuple_prior_tuple.prior.gaussian_tuple_prior_for_arguments(arguments))
        for prior_tuple in self.direct_prior_tuples:
            setattr(new_model, prior_tuple.name, model_arguments[prior_tuple.name])
        for constant_tuple in self.constant_tuples:
            setattr(new_model, constant_tuple.name, constant_tuple.constant)

        return new_model



#!/usr/bin/env python

"""
Filter and collate phase data in all subdirectories.

Usage:

./aggregator.py (root_directory) [pipeline=pipeline phase=phase data=data]

Example:

./aggregator.py ../output pipeline=data_mass_x1_source_x1_positions
"""

import os
import pickle

from autofit.optimize import non_linear


class PhaseOutput(object):
    """
    @DynamicAttrs
    """

    def __init__(self, directory: str):
        """
        Represents the output of a single phase. Comprises a .metadata file and other data files.

        Parameters
        ----------
        directory
            The directory of the phase
        """
        self.directory = directory
        self.__optimizer = None
        self.file_path = os.path.join(directory, ".metadata")
        with open(self.file_path) as f:
            self.text = f.read()
            pairs = [line.split("=") for line in self.text.split("\n")]
            self.__dict__.update({pair[0]: pair[1] for pair in pairs})

    @property
    def model_results(self) -> str:
        """
        Reads the model.results file
        """
        with open(os.path.join(self.directory, "model.results")) as f:
            return f.read()

    @property
    def header(self) -> str:
        """
        A header created by joining the pipeline, phase and data names
        """
        return "/".join((self.pipeline, self.phase, self.data))

    @property
    def optimizer(self) -> non_linear.NonLinearOptimizer:
        """
        The optimizer object that was used in this phase
        """
        if self.__optimizer is None:
            with open(os.path.join(self.directory, ".optimizer.pickle"), "r+b") as f:
                self.__optimizer = pickle.loads(f.read())
        return self.__optimizer

    def __str__(self):
        return self.text

    def __repr__(self):
        return "<PhaseOutput {}>".format(self)


class Aggregator(object):
    def __init__(self, directory: str):
        """
        Class to aggregate phase results for all subdirectories in a given directory.

        The whole directory structure is traversed and a Phase object created for each directory that contains a
        .metadata file.

        Parameters
        ----------
        directory
        """
        self.directory = directory
        self.phases = []

        for root, _, filenames in os.walk(directory):
            if ".metadata" in filenames:
                self.phases.append(PhaseOutput(root))

    def phases_with(self, **kwargs) -> [PhaseOutput]:
        """
        Filters phases. If no arguments are passed all phases are returned. Arguments must be key value pairs, with
        phase, data or pipeline as the key.

        Parameters
        ----------
        kwargs
            Filters, e.g. pipeline=pipeline1
        """
        return [phase for phase in self.phases if
                all([getattr(phase, key) == value for key, value in kwargs.items()])]

    def optimizers_with(self, **kwargs) -> [non_linear.NonLinearOptimizer]:
        """
        Load a list of optimizers for phases in the directory with zero or more filters applied.

        Parameters
        ----------
        kwargs
            Filters, e.g. pipeline=pipeline1

        Returns
        -------
        optimizers
            A list of optimizers, one for each phase in the directory that matches the filters.
        """
        return [phase.optimizer for phase in self.phases_with(**kwargs)]

    def model_results(self, **kwargs) -> str:
        """
        Collates model results from all phases in the directory or some subset if filters are applied.

        Parameters
        ----------
        kwargs
            Filters, e.g. pipeline=pipeline1

        Returns
        -------
        model_results
            A string joining headers and results for all included phases.
        """
        return "\n\n".join("{}\n\n{}".format(phase.header, phase.model_results) for phase in
                           self.phases_with(**kwargs))


if __name__ == "__main__":
    from sys import argv

    root_directory = None
    try:
        root_directory = argv[1]
    except IndexError:
        print("Usage:\n\naggregator.py (root_directory) [pipeline=pipeline phase=phase data=data]")
        exit(1)
    filter_dict = {pair[0]: pair[1] for pair in [line.split("=") for line in argv[2:]]}

    with open("model.results", "w+") as out:
        out.write(Aggregator(root_directory).model_results(**filter_dict))



import autofit.mapper.prior_model
from autofit.mapper import model_mapper as mm, prior as p


class Galaxy(object):
    def __init__(self, redshift=None):
        self.redshift = redshift


class RelativeWidth(object):
    def __init__(self, one, two, three):
        self.one = one
        self.two = two
        self.three = three


class Redshift(object):
    def __init__(self, redshift):
        self.redshift = redshift


# noinspection PyAbstractClass
class GalaxyModel(autofit.mapper.prior_model.AbstractPriorModel):
    def instance_for_arguments(self, arguments):
        try:
            return Galaxy(redshift=self.redshift.instance_for_arguments(arguments))
        except AttributeError:
            return Galaxy()

    def __init__(self, variable_redshift=False, **kwargs):
        self.redshift = autofit.mapper.prior_model.PriorModel(Redshift) if variable_redshift else None
        self.kwargs = kwargs

    @property
    @p.cast_collection(p.PriorNameValue)
    def prior_tuples(self):
        return [item for item in self.__dict__.items() if isinstance(item[1], p.Prior)] + [
            ("redshift", self.redshift.redshift)] if self.redshift is not None else []

    @property
    @p.cast_collection(autofit.mapper.prior_model.PriorModelNameValue)
    def flat_prior_model_tuples(self):
        return [item for item in self.__dict__.items() if isinstance(item[1],
                                                                     autofit.mapper.prior_model.AbstractPriorModel)]


class GeometryProfile(object):

    def __init__(self, centre=(0.0, 0.0)):
        """Abstract GeometryProfile, describing an object with y, x cartesian coordinates"""
        self.centre = centre


class SphericalProfile(GeometryProfile):

    def __init__(self, centre=(0.0, 0.0)):
        """ Generic circular profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profile.
        """
        super(SphericalProfile, self).__init__(centre)


class EllipticalProfile(SphericalProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """ Generic elliptical profiles class to contain functions shared by light and mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalProfile, self).__init__(centre)
        self.axis_ratio = axis_ratio
        self.phi = phi


class EllipticalLP(EllipticalProfile):
    """Generic class for an elliptical light profiles"""

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """  Abstract class for an elliptical light-profile.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalLP, self).__init__(centre, axis_ratio, phi)


class AbstractEllipticalSersic(EllipticalProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0):
        """ Abstract base class for an elliptical Sersic profile, used for computing its effective radius and
        Sersic constant.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) coordinates of the origin of the profiles
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a)
        phi : float
            Rotational angle of profiles ellipse counter-clockwise from positive x-axis
        intensity : float
            Overall intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this model_mapper
        sersic_index : Int
            The Sersic index, which controls the light profile concentration
        """
        super(AbstractEllipticalSersic, self).__init__(centre, axis_ratio, phi)
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index


class MassProfile(object):

    def surface_density_func(self, eta):
        raise NotImplementedError("surface_density_at_radius should be overridden")

    def surface_density_from_grid(self, grid):
        pass
        # raise NotImplementedError("surface_density_from_grid should be overridden")

    def potential_from_grid(self, grid):
        pass
        # raise NotImplementedError("potential_from_grid should be overridden")

    def deflections_from_grid(self, grid):
        raise NotImplementedError("deflections_from_grid should be overridden")

    def dimensionless_mass_within_ellipse(self, major_axis):
        raise NotImplementedError()


# noinspection PyAbstractClass
class EllipticalMassProfile(EllipticalProfile, MassProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0):
        """
        Abstract class for elliptical mass profiles.

        Parameters
        ----------
        centre: (float, float)
            The origin of the profile
        axis_ratio : float
            Ellipse's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of profile's ellipse counter-clockwise from positive x-axis
        """
        super(EllipticalMassProfile, self).__init__(centre, axis_ratio, phi)
        self.axis_ratio = axis_ratio
        self.phi = phi


class EllipticalCoredPowerLaw(EllipticalMassProfile, MassProfile):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, einstein_radius=1.0, slope=2.0, core_radius=0.01):
        """
        Represents a cored elliptical power-law density distribution

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the origin of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        slope : float
            power-law density slope of mass profiles
        core_radius : float
            The radius of the inner core
        """
        super(EllipticalCoredPowerLaw, self).__init__(centre, axis_ratio, phi)
        self.einstein_radius = einstein_radius
        self.slope = slope
        self.core_radius = core_radius


class EllipticalCoredIsothermal(EllipticalCoredPowerLaw):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, einstein_radius=1.0, core_radius=0.05):
        """
        Represents a cored elliptical isothermal density distribution, which is equivalent to the elliptical power-law
        density distribution for the value slope=2.0

        Parameters
        ----------
        centre: (float, float)
            The image_grid of the origin of the profiles
        axis_ratio : float
            Elliptical mass profile's minor-to-major axis ratio (b/a)
        phi : float
            Rotation angle of mass profile's ellipse counter-clockwise from positive x-axis
        einstein_radius : float
            Einstein radius of power-law mass profiles
        core_radius : float
            The radius of the inner core
        """

        super(EllipticalCoredIsothermal, self).__init__(centre, axis_ratio, phi, einstein_radius, 2.0,
                                                        core_radius)


class EllipticalSersic(AbstractEllipticalSersic, EllipticalLP):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0):
        """ The elliptical Sersic profile, used for fitting a model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the light profile.
        """
        super(EllipticalSersic, self).__init__(centre, axis_ratio, phi, intensity, effective_radius,
                                               sersic_index)


class EllipticalCoreSersic(EllipticalSersic):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6,
                 sersic_index=4.0, radius_break=0.01, intensity_break=0.05, gamma=0.25, alpha=3.0):
        """ The elliptical cored-Sersic profile, used for fitting a model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concetration of the of the light profile.
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super(EllipticalCoreSersic, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, sersic_index,
                                                   )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma


class EllipticalExponential(EllipticalSersic):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, effective_radius=0.6):
        """ The elliptical exponential profile, used for fitting a model_galaxy's light.

        This is a subset of the elliptical Sersic profile, specific to the case that sersic_index = 1.0.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        """
        super(EllipticalExponential, self).__init__(centre, axis_ratio, phi, intensity, effective_radius, 1.0)


class EllipticalGaussian(EllipticalLP):

    def __init__(self, centre=(0.0, 0.0), axis_ratio=1.0, phi=0.0, intensity=0.1, sigma=0.01):
        """ The elliptical Gaussian profile.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) origin of the light profile.
        axis_ratio : float
            Ratio of light profiles ellipse's minor and major axes (b/a).
        phi : float
            Rotation angle of light profile counter-clockwise from positive x-axis.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        sigma : float
            The full-width half-maximum of the Gaussian.
        """
        super(EllipticalGaussian, self).__init__(centre, axis_ratio, phi)

        self.intensity = intensity
        self.sigma = sigma
